import logging

import h5py
import numpy as np

from ada.core.containers import Nodes
from ada.fem import FEM, Elem, FemSet
from ada.fem.containers import FemElements, FemSets

from .common import med_to_abaqus_type


def read_fem(assembly, fem_file, fem_name=None):
    """

    :param assembly:
    :param fem_file:
    :param fem_name:
    :return:
    """
    from ada import Node, Part

    f = h5py.File(fem_file, "r")

    # Mesh ensemble
    mesh_ensemble = f["ENS_MAA"]
    meshes = mesh_ensemble.keys()
    if len(meshes) != 1:
        raise ValueError("Must only contain exactly 1 mesh, found {}.".format(len(meshes)))
    mesh_name = list(meshes)[0]
    mesh = mesh_ensemble[mesh_name]

    dim = mesh.attrs["ESP"]

    fem_name = fem_name if fem_name is not None else mesh_name

    # Initialize FEM object
    fem = FEM(mesh_name)

    # Possible time-stepping
    if "NOE" not in mesh:
        # One needs NOE (node) and MAI (French maillage, meshing) data. If they
        # are not available in the mesh, check for time-steppings.
        time_step = mesh.keys()
        if len(time_step) != 1:
            raise ValueError(f"Must only contain exactly 1 time-step, found {len(time_step)}.")
        mesh = mesh[list(time_step)[0]]

    # Points
    pts_dataset = mesh["NOE"]["COO"]
    n_points = pts_dataset.attrs["NBR"]
    points = pts_dataset[()].reshape((n_points, dim), order="F")

    if "NUM" in mesh["NOE"]:
        point_num = list(mesh["NOE"]["NUM"])
    else:
        logging.warning("No node information is found on MED file")
        point_num = np.arange(1, len(points) + 1)

    fem._nodes = Nodes([Node(p, point_num[i]) for i, p in enumerate(points)], parent=fem)

    # Point tags
    tags = None
    if "FAM" in mesh["NOE"]:
        tags = mesh["NOE"]["FAM"][()]

    # Information for point tags
    point_tags = {}
    fas = mesh["FAS"] if "FAS" in mesh else f["FAS"][mesh_name]
    if "NOEUD" in fas:
        point_tags = _read_families(fas["NOEUD"])

    point_sets = _point_tags_to_sets(tags, point_tags, fem) if tags is not None else []

    # Information for cell tags
    cell_tags = {}
    if "ELEME" in fas:
        cell_tags = _read_families(fas["ELEME"])

    # CellBlock
    cell_types = []
    med_cells = mesh["MAI"]

    elements = []
    element_sets = dict()
    for med_cell_type, med_cell_type_group in med_cells.items():
        if med_cell_type == "PO1":
            logging.warning("Point elements are still not supported")
            continue

        cell_type = med_to_abaqus_type(med_cell_type)
        cell_types.append(cell_type)

        nod = med_cell_type_group["NOD"]
        n_cells = nod.attrs["NBR"]
        nodes_in = nod[()].reshape(n_cells, -1, order="F")

        if "NUM" in med_cell_type_group.keys():
            num = list(med_cell_type_group["NUM"])
        else:
            num = np.arange(0, len(nodes_in))

        element_block = [
            Elem(num[i], [fem.nodes.from_id(e) for e in c], cell_type, parent=fem) for i, c in enumerate(nodes_in)
        ]
        elements += element_block
        # Cell tags
        if "FAM" in med_cell_type_group:
            cell_data = med_cell_type_group["FAM"][()]
            cell_type_sets = _cell_tag_to_set(cell_data, cell_tags)
            for key, val in cell_type_sets.items():
                if key not in element_sets.keys():
                    element_sets[key] = []
                element_sets[key] += [element_block[i] for i in val]

    fem._elements = FemElements(elements, fem)

    elsets = []
    for name, values in element_sets.items():
        elsets.append(FemSet(name, values, "elset", parent=fem))

    fem._sets = FemSets(elsets + point_sets, fem_obj=fem)

    assembly.add_part(Part(fem_name, fem=fem))
    return


def _cell_tag_to_set(cell_data_array, cell_tags):
    """
    For a single element type convert tag data into set data

    :param cell_data_array:
    :param cell_tags:
    :return: Cell Sets dictionary
    """
    cell_sets = dict()
    shared_sets = []
    for tag_id, tag_names in cell_tags.items():
        if len(tag_names) > 1:
            for v in tag_names:
                res = np.where(cell_data_array == tag_id)[0]
                if len(res) > 0:
                    shared_sets.append((v, res))
        else:
            tag_name = tag_names[0]
            res = np.where(cell_data_array == tag_id)[0]
            if len(res) > 0:
                cell_sets[tag_name] = res

    for v, s in shared_sets:
        if v in cell_sets.keys():
            cell_sets[v] = np.concatenate([cell_sets[v], s])
        else:
            cell_sets[v] = s

    return cell_sets


def _point_tags_to_sets(tags, point_tags, fem):
    """

    :param tags:
    :param point_tags:
    :return: Point sets dictionary
    """
    point_sets = dict()
    shared_sets = []
    for key, val in point_tags.items():
        if len(val) > 1:
            for set_name in val:
                shared_sets.append((set_name, np.where(tags == key)[0]))
        else:
            point_sets[val[0]] = np.where(tags == key)[0]

    for set_name, s in shared_sets:
        point_sets[set_name] = np.concatenate([point_sets[set_name], s])

    nsets = [FemSet(pn, [fem.nodes.from_id(i + 1) for i in ps], "nset", parent=fem) for pn, ps in point_sets.items()]
    return nsets


def _read_nodal_data(med_data, profiles):
    profile = med_data["NOE"].attrs["PFL"]
    data_profile = med_data["NOE"][profile]
    n_points = data_profile.attrs["NBR"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, -1, order="F")
        values = np.full((n_points, values_profile.shape[1]), np.nan)
        values[index_profile] = values_profile
    if values.shape[-1] == 1:  # cut off for scalars
        values = values[:, 0]
    return values


def _read_cell_data(med_data, profiles):
    profile = med_data.attrs["PFL"]
    data_profile = med_data[profile]
    n_cells = data_profile.attrs["NBR"]
    n_gauss_points = data_profile.attrs["NGA"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_cells, n_gauss_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, n_gauss_points, -1, order="F")
        values = np.full((n_cells, values_profile.shape[1], values_profile.shape[2]), np.nan)
        values[index_profile] = values_profile

    # Only 1 data point per cell, shape -> (n_cells, n_components)
    if n_gauss_points == 1:
        values = values[:, 0, :]
        if values.shape[-1] == 1:  # cut off for scalars
            values = values[:, 0]
    return values


def _read_families(fas_data):
    families = {}
    for _, node_set in fas_data.items():
        set_id = node_set.attrs["NUM"]  # unique set id
        n_subsets = node_set["GRO"].attrs["NBR"]  # number of subsets
        nom_dataset = node_set["GRO"]["NOM"][()]  # (n_subsets, 80) of int8
        name = [None] * n_subsets
        for i in range(n_subsets):
            name[i] = "".join([chr(x) for x in nom_dataset[i]]).strip().rstrip("\x00")
        families[set_id] = name
    return families



import pathlib
from xml.etree import ElementTree as ET

import numpy
from meshio.xdmf.common import CellBlock, translate_mixed_cells, xdmf_to_meshio_type

from .common import ReadError, cell_data_from_raw, xdmf_to_numpy_type


class XdmfReader:
    def __init__(self, filename):
        self.filename = filename

    def read(self):
        parser = ET.XMLParser()
        tree = ET.parse(self.filename, parser)
        root = tree.getroot()

        if root.tag != "Xdmf":
            raise ReadError()

        version = root.get("Version")

        if version.split(".")[0] != "3":
            raise ReadError(f"Unknown XDMF version {version}.")

        return self.read_xdmf3(root)

    def _read_data_item(self, data_item, root=None):
        import h5py

        reference = data_item.get("Reference")
        if reference:
            xpath = (data_item.text if reference == "XML" else reference).strip()
            if xpath.startswith("/"):
                return self._read_data_item(root.find(".//" + "/".join(xpath.split("/")[2:])), root)
            raise ValueError(f"Can't read XPath {xpath}.")

        dims = [int(d) for d in data_item.get("Dimensions").split()]

        # Actually, `NumberType` is XDMF2 and `DataType` XDMF3, but many files
        # out there use both keys interchangeably.
        if data_item.get("DataType"):
            if data_item.get("NumberType"):
                raise ReadError()
            data_type = data_item.get("DataType")
        elif data_item.get("NumberType"):
            if data_item.get("DataType"):
                raise ReadError()
            data_type = data_item.get("NumberType")
        else:
            # Default, see
            # <http://www.xdmf.org/index.php/XDMF_Model_and_Format#XML_Element_.28Xdmf_ClassName.29_and_Default_XML_Attributes>
            data_type = "Float"

        try:
            precision = data_item.attrib["Precision"]
        except KeyError:
            precision = "4"

        if data_item.get("Format") == "XML":
            return numpy.fromstring(
                data_item.text,
                dtype=xdmf_to_numpy_type[(data_type, precision)],
                sep=" ",
            ).reshape(dims)
        elif data_item.get("Format") == "Binary":
            return numpy.fromfile(data_item.text.strip(), dtype=xdmf_to_numpy_type[(data_type, precision)]).reshape(
                dims
            )
        elif data_item.get("Format") != "HDF":
            fmt = data_item.get("Format")
            raise ReadError(f"Unknown XDMF Format '{fmt}'.")

        info = data_item.text.strip()
        filename, h5path = info.split(":")

        # The HDF5 file path is given with respect to the XDMF (XML) file.
        dirname = pathlib.Path(self.filename).resolve().parent
        full_hdf5_path = dirname / filename

        f = h5py.File(full_hdf5_path, "r")

        # Some files don't contain the leading slash /.
        if h5path[0] == "/":
            h5path = h5path[1:]

        for key in h5path.split("/"):
            f = f[key]
        # `[()]` gives a numpy.ndarray
        return f[()]

    def read_information(self, c_data):
        field_data = {}
        root = ET.fromstring(c_data)
        for child in root:
            str_tag = child.get("key")
            dim = int(child.get("dim"))
            num_tag = int(child.text)
            field_data[str_tag] = numpy.array([num_tag, dim])
        return field_data

    def read_xdmf3(self, root):  # noqa: C901
        domains = list(root)
        if len(domains) != 1:
            raise ReadError()
        domain = domains[0]
        if domain.tag != "Domain":
            raise ReadError()

        grids = list(domain)
        if len(grids) != 1:
            raise ReadError("XDMF reader: Only supports one grid right now.")
        grid = grids[0]
        if grid.tag != "Grid":
            raise ReadError()

        points = None
        cells = []
        point_data = {}
        cell_data_raw = {}
        field_data = {}

        for c in grid:
            if c.tag == "Topology":
                data_items = list(c)
                if len(data_items) != 1:
                    raise ReadError()
                data_item = data_items[0]

                data = self._read_data_item(data_item)

                # The XDMF2 key is `TopologyType`, just `Type` for XDMF3.
                # Allow both.
                if c.get("Type"):
                    if c.get("TopologyType"):
                        raise ReadError()
                    cell_type = c.get("Type")
                else:
                    cell_type = c.get("TopologyType")

                if cell_type == "Mixed":
                    cells = translate_mixed_cells(data)
                else:
                    cells.append(CellBlock(xdmf_to_meshio_type[cell_type], data))

            elif c.tag == "Geometry":
                if c.get("Type"):
                    if c.get("GeometryType"):
                        raise ReadError()
                    geometry_type = c.get("Type")
                else:
                    geometry_type = c.get("GeometryType")

                if geometry_type not in ["XY", "XYZ"]:
                    raise ReadError(f'Illegal geometry type "{geometry_type}".')

                data_items = list(c)
                if len(data_items) != 1:
                    raise ReadError()
                data_item = data_items[0]
                points = self._read_data_item(data_item)

            elif c.tag == "Information":
                c_data = c.text
                if not c_data:
                    raise ReadError()
                field_data = self.read_information(c_data)

            elif c.tag == "Attribute":
                # Don't be too strict here: FEniCS, for example, calls this
                # 'AttributeType'.
                # assert c.attrib['Type'] == 'None'

                data_items = list(c)
                if len(data_items) != 1:
                    raise ReadError()
                data_item = data_items[0]

                data = self._read_data_item(data_item)

                name = c.get("Name")
                if c.get("Center") == "Node":
                    point_data[name] = data
                else:
                    if c.get("Center") != "Cell":
                        raise ReadError()
                    cell_data_raw[name] = data
            else:
                raise ReadError(f"Unknown section '{c.tag}'.")

        cell_data = cell_data_from_raw(cells, cell_data_raw)
        from meshio import Mesh

        return Mesh(
            points,
            cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=field_data,
        )



import inspect

import ipywidgets as widgets
import numpy as np
import plotly.graph_objs as go
import scipy.optimize
from IPython.display import display


class Calibrate(object):
    """


    :param mat_data: Material data sheet object
    :param interpolate: All test dataset arrays are interpolated to have equal length.
    :param num_int: Total number of points employed in the linear interpolation of test strain and stress data.
    :param load_types: Manually override which test result you wish to calibrate for. Default is None which means it
                       will be checked for all available test data ['uniaxial', 'biaxial', 'planar']
    :param incompressible: Does nothing (for now)

    :type mat_data: matdb.classes.MatDataSheet
    """

    def __init__(
        self,
        mat_data,
        num_int=300,
        load_types=None,
        interpolate=True,
        incompressible=True,
    ):
        from ada.core.utils import easy_plotly

        self._mat_data = mat_data
        self._num_int = num_int
        self._load_types = ["uniaxial", "biaxial", "planar"] if load_types is None else load_types
        self._incompressible = incompressible
        self._interpolate = interpolate
        self._params = None
        self._model = None
        self._models_d = dict(Yeoh=dict(model=yeoh, init=[1, 1, 1]), Neo=dict(model=neo_hookean, init=[1]))
        self._selected_models = []
        self._fig = easy_plotly("ADA OpenSim", ([], []), return_widget=True)

    def run(self, model, initial_guess, method="leastsq"):
        """
        Run calibration for a specific model

        :param model: Define which polymer model to calibrate for
        :param initial_guess:
        :param method:
        :return:
        """
        import lmfit

        model_name = model.__name__
        params = [x for x in inspect.getfullargspec(model)[0] if x not in ["strain", "load_type"]]
        params_str = ", ".join([f"{c}" for c in params])
        if len(params) != len(initial_guess):
            raise Exception(
                f'Please include sufficient number of parameters for initial_guess for "{model_name}".\n'
                f"Please input a list containing the required parameters, example: [{params_str}]"
            )

        fit_params = lmfit.Parameters()
        for i, lt in enumerate(self._load_types):
            for param, iguess in zip(params, initial_guess):
                fit_params.add(f"{lt}_{param}", value=iguess)

        if len(self._load_types) > 1:
            for lt in self._load_types[1:]:
                for param in params:
                    fit_params[f"{lt}_{param}"].expr = f"{self._load_types[0]}_{param}"

        # run the global fit to all the data sets
        result = lmfit.minimize(self._calibrate, method=method, params=fit_params, args=(model,))
        self._params = [result.params[x].value for x in result.var_names]
        self._model = model
        self.plot(model, self._params)
        return result

    def _calibrate(self, params, model):
        """

        :param params:
        :param model:
        :return:
        """
        from .utils import S_from_sig, e_from_eps

        eps2e = e_from_eps
        sig2s = S_from_sig
        data = list()
        if "uniaxial" in self._load_types:
            data.append(sig2s(self.uni_eps, self.uni_sig))
        if "biaxial" in self._load_types:
            data.append(sig2s(self.bia_eps, self.bia_sig))
        if "planar" in self._load_types:
            data.append(sig2s(self.pla_eps, self.pla_sig))

        data = np.array(data)

        resid = 0.0 * data[:]

        # Uniaxial
        uni_param = [params[x].value for x in params if "uniaxial" in x]
        if len(uni_param) > 0:
            resid[0, :] = data[0, :] - sig2s(
                eps2e(self.uni_eps),
                model(self.uni_eps, *uni_param, load_type="uniaxial"),
            )

        # Biaxial
        bia_param = [params[x].value for x in params if "biaxial" in x]
        if len(bia_param) > 0:
            resid[1, :] = data[1, :] - sig2s(
                eps2e(self.bia_eps),
                model(self.bia_eps, *bia_param, load_type="biaxial"),
            )

        # Planar
        pla_param = [params[x].value for x in params if "planar" in x]
        if len(pla_param) > 0:
            resid[2, :] = data[2, :] - sig2s(eps2e(self.pla_eps), model(self.pla_eps, *pla_param, load_type="planar"))

        # now flatten this to a 1D array, as minimize() needs
        return resid.flatten()

    def _build_test_data(self, eng_plot=True):
        test_data = self.mat_data.eng_dict if eng_plot else self.mat_data.true_dict
        traces = []
        for key, data in test_data.items():
            traces.append(go.Scatter(name=key, x=data[0], y=data[1]))
        return traces

    def _build_plot(self, model, params, eng_plot=True):
        from .utils import S_from_sig, e_from_eps

        model_name = model.__name__
        eps2e = e_from_eps
        sig2_s = S_from_sig
        traces = list()
        coeff = [x for x in inspect.getfullargspec(model)[0] if x not in ["strain", "load_type"]]
        coeff = "".join([f"{c}:{x:>15.4E}<br>" for c, x in zip(coeff, params)])
        hover_template = "{model_name} coefficients:<br>{coeff}".format(model_name=model_name, coeff=coeff)

        if "uniaxial" in self._load_types:
            uni_fit_sig = model(self.uni_eps, *params, load_type="uniaxial")
            uni_fit_strain = eps2e(self.uni_eps) if eng_plot else self.uni_eps
            uni_fit_stress = sig2_s(eps2e(self.uni_eps), uni_fit_sig) if eng_plot else uni_fit_sig
            traces.append(
                go.Scatter(
                    name=f"{model_name} fit (Uniaxial)",
                    x=uni_fit_strain,
                    y=uni_fit_stress,
                    text=hover_template,
                )
            )
        if "biaxial" in self._load_types:
            bia_fit_sig = model(self.bia_eps, *params, load_type="biaxial")
            bia_fit_strain = eps2e(self.bia_eps) if eng_plot else self.bia_eps
            bia_fit_stress = sig2_s(eps2e(self.bia_eps), bia_fit_sig) if eng_plot else bia_fit_sig
            traces.append(
                go.Scatter(
                    name=f"{model_name} fit (Biaxial)",
                    x=bia_fit_strain,
                    y=bia_fit_stress,
                    text=hover_template,
                )
            )
        if "planar" in self._load_types:
            pla_fit_sig = model(self.pla_eps, *params, load_type="planar")
            pla_fit_strain = eps2e(self.pla_eps) if eng_plot else self.pla_eps
            pla_fit_stress = sig2_s(eps2e(self.pla_eps), pla_fit_sig) if eng_plot else pla_fit_sig
            traces.append(
                go.Scatter(
                    name=f"{model_name} fit (Planar)",
                    x=pla_fit_strain,
                    y=pla_fit_stress,
                    text=hover_template,
                )
            )

        return traces

    def plot(self, model, params, eng_plot=True):
        """

        :param model:
        :param params:
        :param eng_plot: Plot using engineering stress and strain.
        :return:
        """
        model_name = model.__name__
        coeff = [x for x in inspect.getfullargspec(model)[0] if x not in ["strain", "load_type"]]
        name = "".join([f"{c}:{x:>15.4E}<br>" for c, x in zip(coeff, params)])
        annotations = [
            go.layout.Annotation(
                text=f"{model_name} coefficients:<br>{name}",
                align="left",
                showarrow=False,
                xref="paper",
                yref="paper",
                x=0.1,
                y=0.95,
                bordercolor="black",
                borderwidth=1,
            )
        ]
        traces = self._build_test_data(eng_plot)
        traces += self._build_plot(model, params, eng_plot)
        from ada.core.utils import easy_plotly

        return easy_plotly(
            self.mat_data.name + f" ({model_name})",
            ([], []),
            traces=traces,
            mode="lines+markers",
            annotations=annotations,
            xlbl="Engineering Strain",
            ylbl="Engineering Stress [MPa]",
            return_widget=True,
        )

    # region Properties
    @property
    def mat_data(self):
        """

        :return:
        :rtype: matdb.classes.MatDataSheet
        """
        return self._mat_data

    @property
    def uniaxial(self):
        """

        :return:
        :rtype: matdb.classes.TestData
        """
        if self.mat_data is not None:
            if self.mat_data.uniaxial is not None:
                return self.mat_data.uniaxial
        else:
            return None

    @property
    def biaxial(self):
        """

        :return:
        :rtype: matdb.classes.TestData
        """
        from . import MatDataSheet

        if self.mat_data is not None:
            assert isinstance(self.mat_data, MatDataSheet)
            if self.mat_data.biaxial is not None:
                return self.mat_data.biaxial
        else:
            return None

    @property
    def planar(self):
        """

        :return:
        :rtype: matdb.classes.TestData
        """
        from . import MatDataSheet

        if self.mat_data is not None:
            assert isinstance(self.mat_data, MatDataSheet)
            if self.mat_data.planar is not None:
                return self.mat_data.planar
        else:
            return None

    @property
    def uni_eps(self):
        """

        :return: Interpolated or original array of uniaxial strain test data
        """
        if self._interpolate is True:
            return np.linspace(self.uniaxial.eps[0], self.uniaxial.eps[-1], self._num_int)
        else:
            return self.uniaxial.eps

    @property
    def uni_sig(self):
        if self._interpolate is True:
            return np.interp(self.uni_eps, self.uniaxial.eps, self.uniaxial.sig)
        else:
            return self.uniaxial.sig

    @property
    def bia_eps(self):
        """

        :return: Interpolated array of biaxial strain test data
        """
        if self._interpolate is True:
            return np.linspace(self.biaxial.eps[0], self.biaxial.eps[-1], self._num_int)
        else:
            return self.biaxial.eps

    @property
    def bia_sig(self):
        if self._interpolate is True:
            return np.interp(self.bia_eps, self.biaxial.eps, self.biaxial.sig)
        else:
            return self.biaxial.sig

    @property
    def pla_eps(self):
        """

        :return: Interpolated array of planar strain test data
        """
        if self._interpolate is True:
            return np.linspace(self.planar.eps[0], self.planar.eps[-1], self._num_int)
        else:
            return self.planar.eps

    @property
    def pla_sig(self):
        if self._interpolate is True:
            return np.interp(self.pla_eps, self.planar.eps, self.planar.sig)
        else:
            return self.planar.sig

    @property
    def aba_str(self):
        """

        :return: Abaqus input str
        """
        if self._model.__name__ == "yeoh":
            params_str = ", ".join([str(x) for x in self._params])
            return f"""*Material, name=yeoh_mat
*Hyperelastic, yeoh
 {params_str},          0.,          0.,          0."""
        elif self._model.__name__ == "neo_hookean":
            return f"""*Material, name=neo_hookean_mat
*Hyperelastic, neo hooke
 {self._params[0]},0."""
        else:
            print(f'unrecognized model "{self._model.__name__}"')
            return ""

    # endregion

    def _clicked_models(self, p):
        self._selected_models = p["new"]

    def _click_plot(self, p):
        self._fig.data = []
        self._fig.add_traces(self._build_test_data())
        model_name = self._selected_models
        model_names = ""
        for m in model_name:
            model = self._models_d[m]["model"]
            model_names += " " + m
            param = self._models_d[m]["init"]
            self.run(model, param)

            traces = self._build_plot(self._model, self._params)
            self._fig.add_traces(traces)

        self._fig.layout["title"] = "ADA OpenSim: " + self._model.__name__

    def _repr_html_(self):
        from ada.core.utils import easy_plotly

        self._fig = easy_plotly("ADA OpenSim", ([], []), return_widget=True)
        opt_models = list(self._models_d.keys())
        self._selected_models = ["Yeoh"]
        opts_odb = widgets.SelectMultiple(options=opt_models, value=["Yeoh"], disabled=False)
        center = widgets.Button(
            description="plot",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
        )
        opts_odb.observe(self._clicked_models, "value")
        center.on_click(self._click_plot)
        # self._clicked_models({'new': 'Yeoh'})
        display(widgets.VBox([widgets.HBox([widgets.VBox([opts_odb, center])]), self._fig]))


# region Incompressible
def neo_hookean(strain, mu, load_type="uniaxial"):
    """
    Neo-Hookean incompressible uniaxial

    :param strain:
    :param mu: Shear modulus
    :param load_type: Type of loading
    :return:
    """
    lam = np.exp(strain)
    if load_type == "uniaxial":
        return mu * (lam * lam - 1.0 / lam)
    elif load_type == "biaxial":
        return mu * (lam * lam - 1.0 / lam ** 4)
    elif load_type == "planar":
        return mu * (lam * lam - 1.0 / lam ** 2)
    else:
        print(f"unknown load type {load_type}")
        return None


def yeoh(strain, c10, c20, c30, load_type="uniaxial"):
    """
    Yeoh incompressible

    :param strain:
    :param c10:
    :param c20:
    :param c30:
    :param load_type:
    :return:
    """
    lam = np.exp(strain)
    i1 = lam ** 2 + 2.0 / lam
    if load_type == "uniaxial":
        return 2 * (c10 + 2 * c20 * (i1 - 3) + 3 * c30 * (i1 - 3) ** 2) * (lam ** 2 - 1.0 / lam)
    elif load_type == "biaxial":
        return 2 * (c10 + 2 * c20 * (i1 - 3) + 3 * c30 * (i1 - 3) ** 2) * (lam ** 2 - 1.0 / lam ** 4)
    elif load_type == "planar":
        return 2 * (c10 + 2 * c20 * (i1 - 3) + 3 * c30 * (i1 - 3) ** 2) * (lam ** 2 - 1.0 / lam ** 2)
    else:
        print(f"unknown load type {load_type}")
        return None


# endregion

# region Compressible
def NH_3D(stretch, param):
    """
    Neo-Hookean. 3D loading specified by stretches.

    param[0]=mu, param[1]=kappa

    :param stretch:
    :param param:
    :return:
    """
    mu, kappa = param
    F = np.array([[stretch[0], 0, 0], [0, stretch[1], 0], [0, 0, stretch[2]]], dtype="float")
    J = np.linalg.det(F)
    Fstar = J ** (-1 / 3) * F
    bstar = np.dot(Fstar, Fstar.T)
    dev_bstar = bstar - np.trace(bstar) / 3 * np.eye(3)
    return mu / J * dev_bstar + kappa * (J - 1) * np.eye(3)


def Yeoh_3D(stretch, param):
    """Yeoh. 3D loading specified by stretches.
    param: [C10, C20, C30, kappa]. Returns true stress."""
    L1 = stretch[0]
    L2 = stretch[1]
    L3 = stretch[2]
    F = np.array([[L1, 0, 0], [0, L2, 0], [0, 0, L3]], dtype="float")
    J = np.linalg.det(F)
    bstar = J ** (-2.0 / 3.0) * np.dot(F, F.T)
    devbstar = bstar - np.trace(bstar) / 3 * np.eye(3)
    I1s = np.trace(bstar)
    return 2 / J * (param[0] + 2 * param[1] * (I1s - 3) + 3 * param[2] * (I1s - 3) ** 2) * devbstar + param[3] * (
        J - 1
    ) * np.eye(3)


def uniaxial_stress(model, true_strain, params):
    """
    Compressible uniaxial loading. Returns true stress.

    :param model:
    :param true_strain: numpy array
    :param params:
    :return:
    """

    stress = np.zeros(len(true_strain))
    for i in range(len(true_strain)):
        lam1 = np.exp(true_strain[i])

        def calc_s22_abs(x):
            return abs(model([lam1, x, x], params)[1, 1])

        # calc_s22_abs = lambda x: abs(model([lam1, x, x], params)[1, 1])
        # search for transverse stretch that gives S22=0
        x0 = 1.0 / np.sqrt(lam1)
        lam2 = scipy.optimize.fmin(calc_s22_abs, x0=x0, xtol=1e-9, ftol=1e-9, disp=False)
        stress[i] = model([lam1, lam2, lam2], params)[0, 0]
    return stress


def biaxial_stress(model, trueStrainVec, params):
    """
    Compressible biaxial loading. Returns true stress.

    :param model:
    :param trueStrainVec:
    :param params:
    :return:
    """

    stress = np.zeros(len(trueStrainVec))
    for i in range(len(trueStrainVec)):
        lam1 = np.exp(trueStrainVec[i])

        def calc_s33_abs(x):
            abs(model([lam1, lam1, x], params)[2, 2])

        # search for transverse stretch that gives S33=0
        lam3 = scipy.optimize.fmin(calc_s33_abs, x0=1 / np.sqrt(lam1), xtol=1e-9, ftol=1e-9, disp=False)

        stress[i] = model([lam1, lam1, lam3], params)[0, 0]
    return stress


def planar_stress(model, trueStrainVec, params):
    """
    Compressible planar loading. Returns true stress.

    :param model:
    :param trueStrainVec:
    :param params:
    :return:
    """
    stress = np.zeros(len(trueStrainVec))
    for i in range(len(trueStrainVec)):
        lam1 = np.exp(trueStrainVec[i])

        def calc_s33_abs(x):
            abs(model([lam1, 1.0, x], params)[2, 2])

        # search for transverse stretch that gives S33=0
        lam3 = scipy.optimize.fmin(calc_s33_abs, x0=1 / np.sqrt(lam1), xtol=1e-9, ftol=1e-9, disp=False)
        stress[i] = model([lam1, 1.0, lam3], params)[0, 0]
    return stress


# endregion



from bisect import bisect_left
from dataclasses import dataclass
from itertools import chain, groupby
from operator import attrgetter

import numpy as np


@dataclass
class COG:
    p: np.array
    tot_mass: float
    tot_vol: float
    sh_mass: float
    bm_mass: float
    no_mass: float


class FemElements:
    """

    :param elements:
    :param fem_obj:
    :type fem_obj: ada.fem.FEM
    """

    def __init__(self, elements=None, fem_obj=None, from_np_array=None):
        self._fem_obj = fem_obj
        if from_np_array is not None:
            elements = self.elements_from_array(from_np_array)

        self._elements = list(elements) if elements is not None else []
        self._idmap = {e.id: e for e in self._elements} if len(self._elements) > 0 else dict()

        if len(self._elements) > 0:
            self._by_types = groupby(self._elements, key=attrgetter("type", "elset"))
        else:
            self._by_types = dict()

    def renumber(self):
        from ada.core.utils import Counter

        elid = Counter(1)

        def mapid2(el):
            """

            :param el:
            :type el: ada.fem.Elem
            :return:
            """
            el.id = next(elid)

        list(map(mapid2, self._elements))
        self._idmap = {e.id: e for e in self._elements} if len(self._elements) > 0 else dict()

        if len(self._elements) > 0:
            self._by_types = groupby(self._elements, key=attrgetter("type", "elset"))
        else:
            self._by_types = dict()

    def link_nodes(self):
        """
        Link element nodes with the parent fem node collection

        """

        def grab_nodes(elem):
            """

            :param elem:
            :type elem: ada.fem.Elem
            """

            nodes = [self._fem_obj.nodes.from_id(no) for no in elem.nodes if type(no) in (int, np.int32)]
            if len(nodes) != len(elem.nodes):
                raise ValueError("Unable to convert element nodes")
            elem._nodes = nodes

        list(map(grab_nodes, self._elements))

    def _build_sets_from_elsets(self, elset, elem_iter):
        """

        :param elset:
        :param elem_iter:
        """
        from ada.fem import FemSet

        if elset is not None and type(elset) == str:

            def link_elset(elem, elem_set_):
                elem._elset = elem_set_

            elements = [self.from_id(el.id) for el in elem_iter]
            if elset not in self._fem_obj.elsets:
                elem_set = FemSet(elset, elements, "elset", parent=self._fem_obj)
                self._fem_obj.sets.add(elem_set)
                list(map(lambda x: link_elset(x, elem_set), elements))
            else:
                self._fem_obj.elsets[elset].add_members(elements)
                elem_set = self._fem_obj.elsets[elset]
                list(map(lambda x: link_elset(x, elem_set), elements))

    def elements_from_array(self, array):
        """

        :param array: A list of numpy arrays formatted as [[elid, n1, n2,...,ni, elset, eltype], ..]
        :return:
        """
        from ada.fem import Elem

        def to_elem(e):
            nodes = [self._fem_obj.nodes.from_id(n) for n in e[3:] if (n == 0 or np.isnan(n)) is False]
            return Elem(e[0], nodes, e[1], e[2], parent=self._fem_obj)

        return list(map(to_elem, array))

    def build_sets(self):
        """
        Create sets from attached elset attribute on elements
        """
        for elset, elements in groupby(self._elements, key=attrgetter("elset")):
            if elset is None:
                continue
            self._build_sets_from_elsets(elset, elements)

    def remove_elements_by_set(self, elset):
        """
        Remove elements from element set. Will remove element set on completion

        :param elset: Pass in a
        :type elset: ada.fem.FemSet
        """
        for el in elset.members:
            self._idmap.pop(el.id)
        self._elements = list(self._idmap.values())
        self._by_types = groupby(self._elements, key=attrgetter("type", "elset"))
        p = elset.parent
        p.sets.elements.pop(elset.name)

    def remove_elements_by_id(self, ids):
        """
        Remove elements from element ids. Will remove elements on completion

        :param ids: Pass in a
        :type ids: int or list of ints
        """
        from collections.abc import Iterable

        ids = list(ids) if isinstance(ids, Iterable) else [ids]
        for id in ids:
            if id in self._idmap.keys():
                self._idmap.pop(id)
        self._elements = list(self._idmap.values())
        self._by_types = groupby(self._elements, key=attrgetter("type", "elset"))

    def __contains__(self, item):
        return item.id in self._idmap.keys()

    def __len__(self):
        return len(self._elements)

    def __iter__(self):
        return iter(self._elements)

    def __getitem__(self, index):
        result = self._elements[index]
        return FemElements(result) if isinstance(index, slice) else result

    def __add__(self, other):
        return FemElements(chain.from_iterable([self.elements, other.elements]), self._fem_obj)

    def __repr__(self):
        data = {}
        for key, val in groupby(self._elements, key=attrgetter("type")):
            if key not in data.keys():
                data[key] = len(list(val))
            else:
                data[key] += len(list(val))
        data_str = ", ".join([f'"{key}": {val}' for key, val in data.items()])
        return f"FemElementsCollection(Elements: {len(self._elements)}, By Type: {data_str})"

    def by_types(self):
        return self._by_types

    def calc_cog(self):
        """
        Calculate COG, total mass and structural Volume of your FEM model based on element mass distributed to element
        nodes

        :return: cogx, cogy, cogz, tot_mass, tot_vol
        :rtype: COG
        """
        from itertools import chain

        from ada.core.utils import (
            global_2_local_nodes,
            normal_to_points_in_plane,
            poly_area,
            unit_vector,
            vector_length,
        )

        def calc_sh_elem(el):
            """

            :param el:
            :type el: ada.Elem
            """
            locz = el.fem_sec.local_z if el.fem_sec.local_z is not None else normal_to_points_in_plane(el.nodes)
            locx = unit_vector(el.nodes[1].p - el.nodes[0].p)
            locy = np.cross(locz, locx)
            origin = el.nodes[0]
            t = el.fem_sec.thickness

            ln = global_2_local_nodes([locx, locy], origin, el.nodes)
            x, y, z = list(zip(*ln))
            area = poly_area(x, y)
            vol_ = t * area
            mass = vol_ * el.fem_sec.material.model.rho
            center = sum([e.p for e in el.nodes]) / len(el.nodes)

            return mass, center, vol_

        def calc_bm_elem(el):
            """

            :param el:
            :type el: ada.Elem
            """
            el.fem_sec.section.properties.calculate()
            nodes_ = el.fem_sec.get_offset_coords()
            elem_len = vector_length(nodes_[-1] - nodes_[0])
            vol_ = el.fem_sec.section.properties.Ax * elem_len
            mass = vol_ * el.fem_sec.material.model.rho
            center = sum([e.p for e in el.nodes]) / len(el.nodes)

            return mass, center, vol_

        def calc_mass_elem(el):
            """

            :param el:
            :type el: ada.Elem
            """
            if el.mass_props.type != "MASS":
                raise NotImplementedError(f'Mass type "{el.mass_props.type}" is not yet implemented')
            mass = el.mass_props.mass
            vol_ = 0.0
            return mass, el.nodes[0].p, vol_

        sh = list(chain(map(calc_sh_elem, self.shell)))
        bm = list(chain(map(calc_bm_elem, self.beams)))
        ma = list(chain(map(calc_mass_elem, self.masses)))

        tot_mass = 0.0
        tot_vol = 0.0
        mcog_ = np.array([0, 0, 0]).astype(float)

        sh_mass = sum([r[0] for r in sh])
        bm_mass = sum([r[0] for r in bm])
        no_mass = sum([r[0] for r in ma])

        for m, c, vol in sh + bm + ma:
            tot_vol += vol
            tot_mass += m
            mcog_ += m * np.array(c).astype(float)

        cog_ = mcog_ / tot_mass

        return COG(cog_, tot_mass, tot_vol, sh_mass, bm_mass, no_mass)

    @property
    def max_el_id(self):
        return max(self._idmap.keys())

    @property
    def elements(self):
        return self._elements

    @property
    def shell(self):
        from ada.fem import ElemShapes

        skipel = ["MASS", "SPRING1"]
        return filter(lambda x: x.type not in skipel and x.type in ElemShapes.shell, self._elements)

    @property
    def beams(self):
        from ada.fem import ElemShapes

        skipel = ["MASS", "SPRING1"]
        return filter(lambda x: x.type not in skipel and x.type in ElemShapes.beam, self._elements)

    @property
    def connectors(self):
        """

        :return: Connector elements (lazy iterator)
        """
        from ada.fem import ElemShapes

        skipel = ["MASS", "SPRING1"]
        return filter(lambda x: x.type not in skipel and x.type in ElemShapes.connectors, self._elements)

    @property
    def masses(self):
        """

        :return: Mass elements (lazy iterator)
        """
        from ada.fem import Mass

        return filter(lambda x: x.type in Mass._valid_types, self._elements)

    @property
    def edges(self):
        """


        :return: List of edges for visualization for all elements
        """

        def grab_nodes(elem):
            return [self._fem_obj.nodes.from_id(elem.nodes[e].id).p for ed_seq in elem.edges_seq for e in ed_seq]

        return list(
            chain.from_iterable(
                map(
                    grab_nodes,
                    filter(lambda x: x.type not in ["MASS", "SPRING1"], self._elements),
                )
            )
        )

    @property
    def edges_alt(self):
        """

        :return: List of edges for visualization for all elements (alternative implementation)
        """

        def grab_ids(elem):
            return tuple([elem.nodes[e].id for ed_seq in elem.edges_seq for e in ed_seq])

        return map(
            grab_ids,
            filter(lambda x: x.type not in ["MASS", "SPRING1"], self._elements),
        )

    def from_id(self, el_id):
        """

        :param el_id:
        :rtype: ada.fem.Elem
        """
        if el_id not in self._idmap.keys():
            raise ValueError(f'The elem id "{el_id}" is not found')
        else:
            return self._idmap[el_id]

    def filter_elements(self, keep_elem=None, delete_elem=None):
        """
        Filter which element types you wish to keep using the "keep_elem" list of elem_types or which elements you wish
        to delete using the 'delete_elem' option.

        :param keep_elem:
        :param delete_elem:
        """
        keep_elem = [el_.lower() for el_ in keep_elem] if keep_elem is not None else None
        delete_elem = [el_.lower() for el_ in delete_elem] if delete_elem is not None else None

        def eval_elem(el):
            if keep_elem is not None:
                return True if el.type.lower() in keep_elem else False
            else:
                return False if el.type.lower() in delete_elem else True

        self._elements = list(filter(eval_elem, self._elements))
        self._by_types = dict(groupby(self._elements, key=attrgetter("type")))
        self._idmap = {e.id: e for e in self._elements}

    @property
    def idmap(self):
        """

        :return:
        """
        return self._idmap

    def add(self, elem):
        """

        :param elem:
        :type elem: ada.fem.Elem
        :return:
        """
        if elem.id is None:
            if len(self._elements) > 0:
                elem._el_id = self._elements[-1].id + 1
            else:
                elem._el_id = 1
        if elem.id in self.idmap.keys():
            raise ValueError(f'Elem id "{elem.id}" already exists or is not set.')

        if elem.parent is None:
            elem.parent = self._fem_obj

        self._elements.append(elem)
        self._idmap[elem.id] = elem

        self._by_types = groupby(self._elements, key=attrgetter("type", "elset"))


class FemSections:
    """

    :param sections:
    :param fem_obj:
    :type fem_obj: ada.fem.FEM
    """

    def __init__(self, sections=None, fem_obj=None):
        self._fem_obj = fem_obj
        self._sections = list(sections) if sections is not None else []
        by_types = self._groupby()
        self._beams = by_types["beams"]
        self._shells = by_types["shells"]
        self._solids = by_types["solids"]
        self._dmap = {e.name: e for e in self._sections} if len(self._sections) > 0 else dict()
        if len(self._sections) > 0:
            self._link_data()

    def _map_materials(self, fem_sec, mat_repo):
        """

        :param fem_sec:
        :type fem_sec: ada.fem.FemSection
        :param mat_repo:
        :type mat_repo: ada.core.containers.Materials
        """

        if type(fem_sec.material) is str:
            fem_sec._material = mat_repo.get_by_name(fem_sec.material)

    def _map_elsets(self, fem_sec, elset_repo):
        """

        :param fem_sec:
        :type fem_sec: ada.fem.FemSection
        """
        from ada.fem import FemSection

        if type(fem_sec.elset) is str:
            if fem_sec.elset not in elset_repo:
                raise ValueError(f'The element set "{fem_sec.elset}" is not imported. ')
            fem_sec._elset = elset_repo[fem_sec.elset]
        elif type(fem_sec) is FemSection:
            pass
        else:
            raise ValueError("Invalid element set has been attached to this fem section")

    def _link_data(self):
        from functools import partial

        mat_repo = self._fem_obj.parent.get_assembly().materials
        list(map(partial(self._map_materials, mat_repo=mat_repo), self._sections))

        elsets_repo = self._fem_obj.elsets
        list(map(partial(self._map_elsets, elset_repo=elsets_repo), self._sections))
        [fem_sec.link_elements() for fem_sec in self._sections]

    def _groupby(self):
        return dict(
            beams=list(filter(lambda x: x.type == "beam", self._sections)),
            shells=list(filter(lambda x: x.type == "shell", self._sections)),
            solids=list(filter(lambda x: x.type == "solid", self._sections)),
        )

    def __contains__(self, item):
        """

        :param item:
        :type item: ada.fem.FemSection
        """
        return item in self._sections

    def __len__(self):
        return len(self._sections)

    def __iter__(self):
        return iter(self._sections)

    def __getitem__(self, index):
        result = self._sections[index]
        return FemSections(result) if isinstance(index, slice) else result

    def __add__(self, other):
        return FemSections(chain(self._sections, other._sections))

    def __repr__(self):
        return (
            f"FemSectionsCollection(Beams: {len(self.beams)}, Shells: {len(self.shells)}, Solids: {len(self.solids)})"
        )

    @property
    def beams(self):
        return self._beams

    @property
    def shells(self):
        return self._shells

    @property
    def solids(self):
        return self._solids

    def edges(self):
        return None

    @property
    def dmap(self):
        return self._dmap

    def index(self, item):
        index = bisect_left(self._sections, item)
        if (index != len(self._sections)) and (self._sections[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def _map_femsec_to_elem(self, elem, fem_sec):
        """

        :param elem:
        :type elem: ada.fem.Elem
        """
        elem.fem_sec = fem_sec

    def add(self, sec):
        """

        :param sec:
        :type sec: ada.fem.FemSection
        :return:
        """
        from functools import partial

        if sec.name in self.dmap.keys() or sec.name is None:
            raise ValueError(f'Section name "{sec.name}" already exists')
            # sec._name = sec.name+'_1'

        self._sections.append(sec)
        if sec.type == "beam":
            self._beams.append(sec)
        elif sec.type == "shell":
            self._shells.append(sec)
        else:
            self._solids.append(sec)

        list(map(partial(self._map_femsec_to_elem, fem_sec=sec), sec.elset.members))
        self._dmap[sec.name] = sec


class FemSets:
    """

    :param sets:
    :param fem_obj:
    :type fem_obj: ada.fem.FEM
    """

    def __init__(self, sets=None, fem_obj=None):
        self._fem_obj = fem_obj
        self._sets = sorted(sets, key=attrgetter("type", "name")) if sets is not None else []
        # Merge same name sets
        self._nomap = self._assemble_sets(self.is_nset) if len(self._sets) > 0 else dict()
        self._elmap = self._assemble_sets(self.is_elset) if len(self._sets) > 0 else dict()
        self._same_names = dict()
        if len(self._sets) > 0:
            self.link_data()

    def add_references(self):
        """
        Add reference to the containing FemSet for each member (node or element)

        :return:
        """

        def _map_ref(el, fem_set):
            el.refs.append(fem_set)

        for _set in self._sets:
            [_map_ref(m, _set) for m in _set.members]

    @staticmethod
    def is_nset(fs):
        return True if fs.type == "nset" else False

    @staticmethod
    def is_elset(fs):
        return True if fs.type == "elset" else False

    def _instantiate_all_members(self, fem_set):
        """

        :param fem_set:
        :type fem_set: ada.fem.FemSet
        """
        from ada import Node
        from ada.fem import Elem

        def get_nset(nref):
            if type(nref) is Node:
                return nref
            else:
                return fem_set.parent.nodes.from_id(nref)

        def get_elset(elref):
            if type(elref) in (int, np.int32):
                return fem_set.parent.elements.from_id(elref)
            elif type(elref) is Elem:
                if elref not in elref.parent.elements and len(elref.parent.elements) != 0:
                    raise ValueError("Element might be doubly defined")
                else:
                    return elref
            else:
                raise ValueError("Elref is not recognized")

        if "generate" in fem_set.metadata.keys():
            if fem_set.metadata["generate"] is True and len(fem_set.members) == 0:
                gen_mem = fem_set.metadata["gen_mem"]
                fem_set._members = [i for i in range(gen_mem[0], gen_mem[1] + 1, gen_mem[2])]
                fem_set.metadata["generate"] = False

        if fem_set.type == "nset":
            if len(fem_set.members) == 1 and type(fem_set.members[0]) is str and type(fem_set.members[0]) is not Node:
                fem_set._members = self.nodes[fem_set.members[0]]
            else:
                fem_set._members = list(map(get_nset, fem_set.members))
            fem_set.parent = self._fem_obj
        else:
            fem_set._members = list(map(get_elset, fem_set.members))
            fem_set.parent = self._fem_obj

        return fem_set

    def link_data(self):
        list(map(self._instantiate_all_members, self._sets))

    def _assemble_sets(self, set_type):
        elsets = dict()
        for elset in filter(set_type, self._sets):
            if elset.name not in elsets.keys():
                elsets[elset.name] = elset
            else:
                self._instantiate_all_members(elset)
                elsets[elset.name] += elset
        return elsets

    def __contains__(self, item):
        return item.name in self._elmap.keys()

    def __len__(self):
        return len(self._sets)

    def __iter__(self):
        return iter(self._sets)

    def __getitem__(self, index):
        result = self._sets[index]
        return FemSets(result, fem_obj=self._fem_obj) if isinstance(index, slice) else result

    def __add__(self, other):
        """

        :param other:
        :type other: FemSets
        :return:
        """
        # TODO: make default choice for similar named sets in a global settings class
        for name, _set in other.nodes.items():
            if name in self._nomap.keys():
                raise ValueError("Duplicate node set name. Consider suppressing this error?")
            self.add(_set)
        for name, _set in other.elements.items():
            if name in self._elmap.keys():
                raise ValueError("Duplicate element set name. Consider suppressing this error?")
            self.add(_set)
        return self
        # return FemSetsCollection(chain(self.sets, other.sets), fem_obj=self._fem_obj)

    def get_elset_from_name(self, name):
        """

        :param name:
        :rtype: ada.fem.classes.Elem
        """
        if name not in self._elmap.keys():
            raise ValueError(f'The elem id "{name}" is not found')
        else:
            return self._elmap[name]

    def get_nset_from_name(self, name):
        """

        :param name:
        :rtype: ada.fem.classes.Elem
        """
        if name not in self._nomap.keys():
            raise ValueError(f'The node id "{name}" is not found')
        else:
            return self._nomap[name]

    @property
    def elements(self):
        """

        :return:
        """
        return self._elmap

    @property
    def nodes(self):
        """

        :return:
        """
        return self._nomap

    @property
    def sets(self):
        return self._sets

    def index(self, item):
        index = bisect_left(self._sets, item)
        if (index != len(self._sets)) and (self._sets[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def remove(self, fe_set):
        """

        :param fe_set:
        :type fe_set:
        """
        i = self._sets.index(fe_set)
        self._sets.pop(i)
        if fe_set.type == "nset":
            self._nomap.pop(fe_set.name)
            if fe_set.name in self._same_names.keys():
                self._same_names.pop(fe_set.name)
        else:
            self._elmap.pop(fe_set.name)
            if fe_set.name in self._same_names.keys():
                self._same_names.pop(fe_set.name)

        # To evalute if dependencies of set should be checked?
        # Against: This is a downstream object. FemSections would point to this set and remove during concatenation.

    def add(self, fe_set, append_suffix_on_exist=False):
        """

        :param fe_set:
        :param append_suffix_on_exist:
        :type fe_set: ada.fem.FemSet
        """
        if fe_set.type == "nset":
            if fe_set.name in self._nomap.keys():
                fem_set = self._nomap[fe_set.name]
                new_mem = [m for m in fe_set.members if m.id not in fem_set.members]
                fem_set.add_members(new_mem)
        else:
            if fe_set.name in self._elmap.keys():
                if append_suffix_on_exist is False:
                    raise ValueError("An elements set with the same name already exists")
                if fe_set.name not in self._same_names.keys():
                    self._same_names[fe_set.name] = 1
                else:
                    self._same_names[fe_set.name] += 1

                fe_set._name = f"{fe_set.name}_{self._same_names[fe_set.name]}"

        # if False in list(map(lambda x: type(x) is Node, fe_set.members)):
        #     self._map_members(fe_set)
        self._sets.append(fe_set)
        if fe_set.type == "elset":
            self._elmap[fe_set.name] = fe_set
        else:
            self._nomap[fe_set.name] = fe_set
        if fe_set.parent is None:
            fe_set.parent = self._fem_obj
        self._instantiate_all_members(fe_set)



import math

from ada.core.utils import roundoff as rd


class GeneralProperties:
    def __init__(
        self,
        ax=None,
        ix=None,
        iy=None,
        iz=None,
        iyz=None,
        wxmin=None,
        wymin=None,
        wzmin=None,
        shary=None,
        sharz=None,
        scheny=None,
        schenz=None,
        sy=None,
        sz=None,
        sfy=1,
        sfz=1,
        cy=None,
        cz=None,
        parent=None,
    ):
        self._Ax = ax
        self._Ix = ix
        self._Iy = iy
        self._Iz = iz
        self._Iyz = iyz
        self._Wxmin = wxmin
        self._Wymin = wymin
        self._Wzmin = wzmin
        self._Shary = shary
        self._Sharz = sharz
        self._Scheny = scheny
        self._Schenz = schenz
        self._Sy = sy
        self._Sz = sz
        self._Sfy = sfy
        self._Sfz = sfz
        self._Cy = cy
        self._Cz = cz
        self._parent = parent

    def edit(self, parent):
        self._parent = parent

    def _calc_box(self):
        """
        Calculate box cross section properties
        """

        sfy = 1.0
        sfz = 1.0
        s = self.parent
        self._Ax = s.w_btn * s.t_fbtn + s.w_top * s.t_ftop + s.t_w * (s.h - (s.t_fbtn + s.t_ftop)) * 2
        a = s.t_fbtn
        b = (s.h + s.t_fbtn - s.t_ftop) / 2
        c = s.h - s.t_ftop
        d = s.h - s.t_fbtn - s.t_ftop
        e = s.w_top * s.t_fbtn
        f = s.w_top * s.t_ftop
        g = s.t_w * d
        h = (e * a + f * c + 2 * b * g) / self.Ax
        ha = s.h - (s.t_fbtn + s.t_ftop) / 2.0
        hb = s.w_top - s.t_w

        self._Ix = 4 * (ha * hb) ** 2 / (hb / s.t_fbtn + hb / s.t_ftop + 2 * ha / s.t_w)
        self._Iy = (
            (s.w_top * (s.t_fbtn ** 3 + s.t_ftop ** 3) + 2 * s.t_w * d ** 3) / 12
            + e * (h - a) ** 2
            + f * (c - h) ** 2
            + 2 * g * (b - h) ** 2
        )
        self._Iz = ((s.t_fbtn + s.t_ftop) * s.w_top ** 3 + 2 * d * s.t_w ** 3) / 12 + (g * hb ** 2) / 2
        self._Iyz = 0
        self._Wxmin = self.Ix * (hb + ha) / (ha * hb)
        self._Wymin = self.Iy / max(s.h - h, h)
        self._Wzmin = 2 * self._Iz / s.w_top
        self._Sy = e * (h - a) + s.t_w * (h - s.t_fbtn) ** 2
        self._Sz = (s.t_fbtn + s.t_ftop) * s.w_top ** 2 / 8 + g * hb / 2
        self._Shary = (self.Iz / self.Sz) * 2 * s.t_w * sfy
        self._Sharz = (self.Iy / self.Sy) * 2 * s.t_w * sfz
        self._Scheny = 0
        self._Schenz = c - h - s.t_fbtn * ha / (s.t_fbtn + s.t_ftop)
        self._Cy = s.w_top / 2
        self._Cz = h

    def _calc_isec(self):
        """
        Calculate I/H cross section properties
        """

        sfy = 1.0
        sfz = 1.0
        s = self.parent
        hz = s.h
        bt = s.w_top
        tt = s.t_ftop
        ty = s.t_w
        bb = s.w_btn
        tb = s.t_fbtn

        self._Ax = bt * tt + ty * (hz - (tb + tt)) + bb * tb
        hw = hz - tt - tb
        a = tb + hw + tt / 2
        b = tb + hw / 2
        c = tb / 2

        z = (tb * tt * a + hw * ty * b + bb * tb * c) / self.Ax
        tra = (bb * tb ** 3) / 12 + bt * tt * (hz - tt / 2 - z) ** 2
        trb = (ty * hw ** 3) / 12 + ty * hw * (tb + hw / 2 - z) ** 2
        trc = (bb * tb ** 3) / 12 + bb * tb * (tb / 2 - z) ** 2

        if tt == ty and tt == tb:
            self._Ix = (tt ** 3) * (hw + bt + bb - 1.2 * tt) / 3
            self._Wxmin = self.Ix / tt
        else:
            self._Ix = 1.3 * (bt * tt ** 3 + hw * ty ** 3 + bb * tb ** 3) / 3
            self._Wxmin = self.Ix / max(tt, ty, tb)

        self._Iy = tra + trb + trc
        self._Iz = (tb * bb ** 3 + hw * ty ** 3 + tt * bt ** 3) / 12
        self._Iyz = 0
        self._Wymin = self.Iy / max(hz - z, z)
        self._Wzmin = 2 * self.Iz / max(bb, bt)

        # Sy should be checked. Confer older method implementation.
        self._Sy = (((tt * bt) ** 2) * (hw / 2 + tt / 2)) * 2
        self._Sz = (tt * bt ** 2 + tb * bb ** 2 + hw * ty ** 2) / 8
        self._Shary = (self.Iz / self.Sz) * (tb + tt) * sfy
        self._Sharz = (self.Iy / self.Sy) * ty * sfz
        self._Scheny = 0
        self._Schenz = ((hz - tt / 2) * tt * bt ** 3 + (tb ** 2) * (bb ** 3) / 2) / (tt * bt ** 3 + tb * bb ** 3) - z
        self._Cy = bb / 2
        self._Cz = z

    def _calc_angular(self):
        """
        Calculate L cross section properties
        """
        s = self.parent
        posweb = True
        hz = s.h
        ty = s.t_w
        tz = s.t_fbtn
        by = s.w_btn
        sfy = 1.0
        sfz = 1.0

        r = 0
        hw = hz - tz
        b = tz - hw / 2
        c = tz / 2
        piqrt = math.atan(1.0)
        self._Ax = ty * hw + by * tz + (1 - piqrt) * r ** 2
        y = (hw * ty ** 2 + tz * by ** 2) / (2 * self.Ax)
        z = (hw * b * ty + tz * by * c) / self.Ax
        d = 6 * r + 2 * (ty + tz - math.sqrt(4 * r * (2 * r + ty + tz) + 2 * ty * tz))
        e = hw + tz - z
        f = hw - e
        ri = y - ty
        rj = by - y
        rk = ri + 0.5 * ty
        rl = z - c

        if tz >= ty:
            h = hw
        else:
            raise ValueError("Currently not implemented this yet")

        self._Iy = (ty * hw ** 3 + by * tz ** 3) / 12 + hw * ty * (b - z) ** 2 + by * tz * (z - c) ** 2
        self._Iz = (hw * ty ** 3 + tz * by ** 3) / 12 + hw * ty * rk ** 2 + tz * by * (by / 2 - y) ** 2
        self._Iyz = (rl * tz / 2) * (y ** 2 - rj ** 2) - (rk * ty / 2) * (e ** 2 - f ** 2)

        # This is incorrect. Should find this in my old calculation method.
        self._Ix = self.Iy + self.Iz

        self._Wxmin = self.Ix / d
        self._Wymin = self.Iy / max(z, hz - h)
        self._Wzmin = self.Iz / max(y, rj)
        self._Sy = (ty * e ** 2) / 2
        self._Sz = (tz * rj ** 2) / 2
        self._Shary = (self.Iz * tz / self.Sz) * sfy
        self._Sharz = (self.Iy * tz / self.Sy) * sfz

        if posweb:
            self._Iyz = -self._Iyz
            self._Scheny = rk
            self._Cy = by - y
        else:
            self._Scheny = -rk
            self._Cy = y
        self._Schenz = -rl
        self._Cz = z

    def _calc_tubular(self):
        """
        Calculate Tubular cross section properties
        """
        s = self.parent
        dy = s.r * 2
        t = s.wt
        sfy = 1.0
        sfz = 1.0
        di = dy - 2 * t
        self._Ax = math.pi * s.r ** 2 - math.pi * (s.r - t) ** 2
        self._Ix = 0.5 * math.pi * ((dy / 2) ** 4 - (di / 2) ** 4)
        self._Iy = self.Ix / 2
        self._Iz = self.Iy
        self._Iyz = 0
        self._Wxmin = 2 * self.Ix / dy
        self._Wymin = 2 * self.Iy / dy
        self._Wzmin = 2 * self.Iz / dy
        self._Sy = (dy ** 3 - di ** 3) / 12
        self._Sz = self.Sy
        self._Shary = (2 * self.Iz * t / self.Sy) * sfy
        self._Sharz = (2 * self.Iy * t / self.Sz) * sfz
        self._Scheny = 0
        self._Schenz = 0

    def _calc_circular(self):

        s = self.parent
        self._Ax = math.pi * s.r ** 2
        self._Iy = (math.pi * s.r ** 4) / 4
        self._Iz = self._Iy

    def _calc_flatbar(self):
        s = self.parent
        self._Ax = (s.w_btn + s.w_top) / 2 * s.h
        self._Iy = s.w_top * s.h ** 3 / 12
        self._Iz = s.h * s.w_top ** 3 / 12

    def _calc_channel(self):
        """

        :return:
        """
        posweb = True
        s = self.parent
        hz = s.h
        ty = s.t_w
        tz = s.t_fbtn
        by = s.w_btn
        sfy = 1.0
        sfz = 1.0

        a = hz - 2 * tz
        self._Ax = 2 * by * tz + a * ty
        y = (2 * tz * by ** 2 + a * ty ** 2) / (2 * self.Ax)
        self._Iy = (ty * a ** 3) / 12 + 2 * ((by * tz ** 3) / 12 + by * tz * ((a + tz) / 2) ** 2)

        if tz == ty:
            self._Ix = ty ** 3 * (2 * by + a - 2.6 * ty) / 3
            self._Wxmin = self.Ix / self.Iy
        else:
            self._Ix = 1.12 * (2 * by * tz ** 3 + a * ty ** 3) / 3
            self._Wxmin = self._Ix / max(tz, ty)

        self._Iz = (
            2 * ((tz * by ** 3) / 12 + tz * by * (by / 2 - y) ** 2) + (a * ty ** 3) / 12 + a * ty * (y - ty / 2) ** 2
        )
        self._Iyz = 0
        self._Wymin = 2 * self.Iy / hz
        self._Wzmin = self.Iz / max(by - y, y)
        self._Sy = by * tz * (tz + a) / 2 + (ty * a ** 2) / 8
        self._Sz = tz * (by - y) ** 2
        self._Shary = (self._Iz / self.Sz) * (2 * tz) * sfy
        self._Sharz = (self._Iy / self.Sy) * (2 * ty) * sfz

        if tz == ty:
            q = ((by - ty / 2) ** 2) * ((hz - tz) ** 2) * tz / 4 * self.Iy
        else:
            q = ((by - ty / 2) ** 2) * tz / (2 * (by - ty / 2) * tz + (hz - tz) * ty / 3)

        if posweb:
            self._Scheny = y - ty / 2 + q
            self._Cy = by
        else:
            self._Scheny = -(y - ty / 2 + q)
            self._Cy = by - y

    def calculate(self):
        """
        Calculates the cross section properties based on the parent section.

        A large parts of the calculations are based on the document

            DNVGL. (2011). Appendix B Section properties & consistent units Table of Contents. I.

        Which in turn bases most (if not all) formulas on the work in

            * W. Beitz, K.H. Küttner: "Dubbel, Taschenbuch für den Maschinenbau" 17. Auflage (17th ed.)
              Springer-Verlag 1990
            * Arne Selberg: "Stålkonstruksjoner" Tapir 1972
            * S. Timoshenko: "Strength of Materials, Part I, Elementary Theory and Problems" Third Edition 1995 D.
              Van Nostrand Company Inc.

        """

        if self.parent.type in SectionCat.circular:
            self._calc_circular()
        elif SectionCat.is_i_profile(self.parent.type):
            self._calc_isec()
        elif SectionCat.is_box_profile(self.parent.type):
            self._calc_box()
        elif self.parent.type in SectionCat.general:
            pass  # it is known
        elif self.parent.type in SectionCat.tubular:
            self._calc_tubular()
        elif SectionCat.is_hp_profile(self.parent.type):
            self._calc_angular()
        elif self.parent.type in SectionCat.channels:
            self._calc_channel()
        elif SectionCat.is_flatbar(self.parent.type):
            self._calc_flatbar()
        else:
            raise Exception(
                f'section type "{self.parent.type}" is not yet supported in the cross section parameter calculations'
            )

    @property
    def parent(self):
        """

        :return:
        :rtype: ada.Section
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def Ax(self):
        """

        :return: Area of cross section
        """
        return self._Ax

    @property
    def Ix(self):
        """
        Torsional moment of inertia about shear centre
        :return:
        """
        return self._Ix

    @Ix.setter
    def Ix(self, value):
        if value <= 0.0:
            raise ValueError("Value cannot be zero or non-positive")
        self._Ix = value

    @property
    def Iy(self):
        """

        :return: Moment of inertia about y-axis
        """
        return self._Iy

    @Iy.setter
    def Iy(self, value):
        if value <= 0.0:
            raise ValueError("Value cannot be zero or non-positive")
        self._Iy = value

    @property
    def Iz(self):
        """

        :return: Moment of inertia about z-axis
        """
        return self._Iz

    @Iz.setter
    def Iz(self, value):
        if value <= 0.0:
            raise ValueError("Value cannot be zero or non-positive")
        self._Iz = value

    @property
    def Iyz(self):
        """

        :return: Product of inertia about y- and z-axes
        """
        return self._Iyz

    @Iyz.setter
    def Iyz(self, value):
        if value <= 0.0:
            raise ValueError("Value cannot be zero or non-positive")
        self._Iyz = value

    @property
    def Wxmin(self):
        """

        :return: Minimum torsional sectional modulus about shear centre
        """
        return self._Wxmin

    @property
    def Wymin(self):
        """

        :return: Minimum sectional modulus about y-axis
        """
        return self._Wymin

    @property
    def Wzmin(self):
        """

        :return: Minimum sectional modulus about z-axis
        """

        return self._Wzmin

    @property
    def Shary(self):
        """

        :return: Shear area in the direction of y-axis
        """
        return self._Shary

    @property
    def Sharz(self):
        """

        :return: Shear area in the direction of z-axis
        """
        return self._Sharz

    @property
    def Scheny(self):
        """

        :return: Shear centre location y-component
        """
        return self._Scheny

    @property
    def Schenz(self):
        """

        :return: Shear centre location z-component
        """
        return self._Schenz

    @property
    def Sy(self):
        """

        :return: Static area moment about y-axis
        """
        return self._Sy

    @property
    def Sz(self):
        """

        :return: Static area moment about z-axis
        """
        return self._Sz

    @property
    def Cz(self):
        return self._Cz

    @property
    def Sfy(self):
        """
        :return: Centroid location from bottom right corner y-component
        """
        return self._Sfy

    @property
    def Sfz(self):
        """
        :return: Centroid location from bottom right corner z-component
        """
        return self._Sfz

    def __eq__(self, other):
        for key, val in self.__dict__.items():
            if "parent" in key:
                continue
            if other.__dict__[key] != val:
                return False

        return True


class ProfileBuilder:
    origin = (0.0, 0.0, 0.0)
    """
    A class for creating generalized 2d point curves describing beam section profiles
    """

    @classmethod
    def angular(cls, sec, return_solid):
        h = sec.h
        wbtn = sec.w_btn
        p2 = (0.0, -h)
        p3 = (wbtn, -h)
        disconnected = False
        if return_solid is False:
            disconnected = True
            outer_curve, inner_curve = [(cls.origin[:-1], p2), (p2, p3)], None
        else:
            tf = sec.t_fbtn
            tw = sec.t_w
            p4 = (wbtn, -h + tf)
            p5 = (tw, -h + tf)
            p6 = (tw, 0.0)
            outer_curve, inner_curve = [cls.origin[:-1], p2, p3, p4, p5, p6], None
        return outer_curve, inner_curve, disconnected

    @classmethod
    def iprofiles(cls, sec, return_solid):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :type return_solid:
        """
        h = sec.h
        wbtn = sec.w_btn
        wtop = sec.w_top

        # top flange
        c1 = (-wtop / 2, h / 2)
        c2 = (wtop / 2, h / 2)
        # web
        p3 = (0.0, h / 2)
        p4 = (0.0, -h / 2)
        # bottom flange
        c3 = (-wbtn / 2, -h / 2)
        c4 = (wbtn / 2, -h / 2)

        if return_solid is False:
            outer_curve, inner_curve, disconnected = (
                [(c1, c2), (p3, p4), (c3, c4)],
                None,
                True,
            )
        else:
            tfbtn = sec.t_fbtn
            tftop = sec.t_ftop
            tw = sec.t_w
            p3 = (wtop / 2, h / 2 - tftop)
            p4 = (tw / 2, h / 2 - tftop)
            p5 = (tw / 2, -h / 2 + tfbtn)
            p6 = (wbtn / 2, -h / 2 + tfbtn)
            p7 = (-wbtn / 2, -h / 2 + tfbtn)
            p8 = (-tw / 2, -h / 2 + tfbtn)
            p9 = (-tw / 2, h / 2 - tftop)
            p10 = (-wtop / 2, h / 2 - tftop)
            outer_curve, inner_curve, disconnected = (
                [c1, c2, p3, p4, p5, p6, c4, c3, p7, p8, p9, p10],
                None,
                False,
            )

        return outer_curve, inner_curve, disconnected

    @classmethod
    def box(cls, sec, return_solid):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :type return_solid:
        """
        h = sec.h
        wtop = sec.w_top
        wbtn = sec.w_btn

        p1 = (rd(-wtop / 2), rd(h / 2))
        p2 = (rd(wtop / 2), rd(h / 2))
        p3 = (rd(wbtn / 2), rd(-h / 2))
        p4 = (rd(-wbtn / 2), rd(-h / 2))

        if return_solid is False:
            outer_curve, inner_curve, disconnected = [p1, p2, p3, p4], None, False
        else:
            tftop = sec.t_fbtn
            tfbtn = sec.t_fbtn
            tw = sec.t_w
            p5 = (rd(-wtop / 2 + tw), rd(h / 2 - tftop))
            p6 = (rd(wtop / 2 - tw), rd(h / 2 - tftop))
            p7 = (rd(wbtn / 2 - tw), rd(-h / 2 + tfbtn))
            p8 = (rd(-wbtn / 2 + tw), rd(-h / 2 + tfbtn))

            outer_curve = [p1, p2, p3, p4]
            inner_curve = [p5, p6, p7, p8]
            disconnected = False

        return outer_curve, inner_curve, disconnected

    @classmethod
    def tubular(cls, sec, return_solid):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :type return_solid: bool
        """
        return None, None, None

    @classmethod
    def circular(cls, sec, return_solid=None):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :type return_solid:
        """
        return sec.r, None, False

    @classmethod
    def flatbar(cls, sec, return_solid=None):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :return:
        """
        outer_curve = [
            (-sec.w_top / 2, sec.h / 2),
            (sec.w_top / 2, sec.h / 2),
            (sec.w_top / 2, -sec.h / 2),
            (-sec.w_top / 2, -sec.h / 2),
        ]
        inner_curve = None
        disconnected = False
        return outer_curve, inner_curve, disconnected

    @classmethod
    def gensec(cls, sec, return_solid=None):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :return:
        """
        import numpy as np

        radius = np.sqrt(sec.properties.Ax / np.pi)
        inner_curve = None
        disconnected = False
        return radius, inner_curve, disconnected

    @classmethod
    def channel(cls, sec, return_solid=None):
        """

        :param sec:
        :param return_solid:
        :type sec: ada.Section
        :return:
        """
        outer_curve = [
            (sec.w_top, sec.h / 2 - sec.t_ftop),
            (sec.w_top, sec.h / 2),
            (0, sec.h / 2),
            (0, -sec.h / 2),
            (sec.w_top, -sec.h / 2),
            (sec.w_top, -sec.h / 2 + sec.t_fbtn),
            (sec.t_w, -sec.h / 2 + sec.t_fbtn),
            (sec.t_w, sec.h / 2 - sec.t_fbtn),
        ]
        inner_curve = None
        disconnected = False
        return outer_curve, inner_curve, disconnected

    @staticmethod
    def build_representation(beam, solid=True):
        """
        Build a 3d model by

        :param beam:
        :param solid:
        :type beam: ada.Beam
        :return:
        """
        from OCC.Core.BRepAlgoAPI import BRepAlgoAPI_Cut, BRepAlgoAPI_Fuse
        from OCC.Core.BRepOffsetAPI import BRepOffsetAPI_ThruSections
        from OCC.Core.ShapeUpgrade import ShapeUpgrade_UnifySameDomain
        from OCC.Core.TopoDS import TopoDS_Face, TopoDS_Wire

        from ada.core.utils import face_to_wires, tuple_minus
        from ada.sections import SectionCat

        xdir, ydir, zdir = beam.ori
        ydir_neg = tuple_minus(ydir) if beam.section.type not in SectionCat.angular else tuple(ydir)

        sec = beam.section.cross_sec_shape(
            solid,
            origin=tuple(beam.n1.p.astype(float)),
            xdir=ydir_neg,
            normal=tuple(xdir),
        )
        tap = beam.taper.cross_sec_shape(
            solid,
            origin=tuple(beam.n2.p.astype(float)),
            xdir=ydir_neg,
            normal=tuple(xdir),
        )

        def through_section(sec_a, sec_b, solid_):
            generator_sec = BRepOffsetAPI_ThruSections(solid_, False)
            generator_sec.AddWire(sec_a)
            generator_sec.AddWire(sec_b)
            generator_sec.Build()
            return generator_sec.Shape()

        if type(sec) is TopoDS_Face:
            sec_result = face_to_wires(sec)
            tap_result = face_to_wires(tap)
        elif type(sec) is TopoDS_Wire:
            sec_result = [sec]
            tap_result = [tap]
        else:
            assert isinstance(sec, list)
            sec_result = sec
            tap_result = tap

        shapes = list()
        for s_, t_ in zip(sec_result, tap_result):
            shapes.append(through_section(s_, t_, solid))

        if beam.section.type in SectionCat.box + SectionCat.tubular + SectionCat.rhs + SectionCat.shs and solid is True:
            cut_shape = BRepAlgoAPI_Cut(shapes[0], shapes[1]).Shape()
            shape_upgrade = ShapeUpgrade_UnifySameDomain(cut_shape, False, True, False)
            shape_upgrade.Build()
            return shape_upgrade.Shape()

        if len(shapes) == 1:
            return shapes[0]
        else:
            result = shapes[0]
            for s in shapes[1:]:
                result = BRepAlgoAPI_Fuse(result, s).Shape()
            return result


class SectionCat:
    box = ["BG", "CG"]
    shs = ["SHS"]
    rhs = ["RHS", "URHS"]
    tubular = ["TUB", "PIPE", "OD"]
    iprofiles = ["HEA", "HEB", "IPE"]
    igirders = ["IG"]
    tprofiles = ["TG"]
    angular = ["HP"]
    channels = ["UNP"]
    circular = ["CIRC"]
    general = ["GENBEAM"]
    flatbar = ["FB"]

    @classmethod
    def isbeam(cls, bmtype):
        for key, val in cls.__dict__.items():
            if bmtype in val:
                return True
        return False

    @classmethod
    def is_i_profile(cls, bmtype):
        return True if bmtype.upper() in cls.igirders + cls.iprofiles + cls.tprofiles else False

    @classmethod
    def is_box_profile(cls, bmtype):
        return True if bmtype.upper() in cls.box + cls.shs + cls.rhs else False

    @classmethod
    def is_hp_profile(cls, bmtype):
        return True if bmtype.upper() in cls.angular else False

    @classmethod
    def is_circular_profile(cls, bmtype):
        return True if bmtype.upper() in cls.tubular + cls.circular else False

    @classmethod
    def is_flatbar(cls, bmtype):
        return True if bmtype.upper() in cls.flatbar else False



from ada import CurvePoly, Part, Shape


class Window(Part):
    def __init__(self, name, width, height, depth, **kwargs):
        origin = (0, 0, 0)
        lx = (1, 0, 0)
        ly = (0, 1, 0)
        lz = (0, 0, 1)
        super().__init__(name, origin=origin, lx=lx, ly=ly, lz=lz, **kwargs)
        self._metadata["ifc_type"] = "IfcWindow"
        self._width = width
        self._height = height
        self._depth = depth

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def depth(self):
        return self._depth

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if value != self._units:
            scale_factor = self._unit_conversion(self._units, value)
            self._origin = tuple([x * scale_factor for x in self._origin])
            self._width *= scale_factor
            self._height *= scale_factor
            self._depth *= scale_factor
            self._shapes = []
            self.build_geom()
            self._units = value

    def build_geom(self):
        normal = self._lz
        origin = self.origin - self._lz * self.depth
        points = [(0, 0), (self.width, 0), (self.width, self.height), (0, self.height)]
        poly = CurvePoly(points2d=points, origin=origin, normal=normal, xdir=self._lx, parent=self)
        geom = poly.make_extruded_solid(self.depth)
        self.add_shape(Shape(self.name, geom, metadata=self.metadata))


class Door(Part):
    def __init__(self, name, width, height, depth, units="m", **kwargs):
        origin = (0, 0, 0)
        lx = (1, 0, 0)
        ly = (0, 1, 0)
        lz = (0, 0, 1)
        super().__init__(name, origin=origin, lx=lx, ly=ly, lz=lz, units=units, **kwargs)
        self._metadata["ifc_type"] = "IfcDoor"
        self._width = width
        self._height = height
        self._depth = depth

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def depth(self):
        return self._depth

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if value != self._units:
            scale_factor = self._unit_conversion(self._units, value)
            self._origin = tuple([x * scale_factor for x in self._origin])
            self._width *= scale_factor
            self._height *= scale_factor
            self._depth *= scale_factor
            self._shapes = []
            self.build_geom()
            self._units = value

    def build_geom(self):
        origin = self._origin - self._lz * self.depth

        points = [(0, 0), (self.width, 0), (self.width, self.height), (0, self.height)]
        poly = CurvePoly(points2d=points, origin=origin, normal=self._lz, xdir=self._lx, parent=self)
        geom = poly.make_extruded_solid(self.depth)
        self.add_shape(Shape(self.name, geom, metadata=self.metadata))



import logging

import numpy as np

from ada import Part, PrimBox
from ada.core.containers import Beams
from ada.core.utils import beam_cross_check


def joint_map(name, members):
    """
    :param name: Name of joint
    :param members: Number of members
    :return: Joint
    """

    def mem_type_check(joint):
        mem_types = [m for m in joint.mem_types]

        for m in members:
            if m.member_type in mem_types:
                mem_types.pop(mem_types.index(m.member_type))

        if len(mem_types) != 0:
            return False
        else:
            return True

    def eval_joint_req(joint):
        if len(members) == joint.num_mem and mem_type_check(joint) is True:
            return True
        else:
            return False

    joints = [JointB]

    for joint in joints:
        if eval_joint_req(joint):
            return joint(name, members)
    else:
        member_types = [m.section.type for m in members]
        logging.error(f'Unable to find matching Joint using joint map for members "{member_types}"')
        return None


class JointBase(Part):
    beamtypes: list
    mem_types: list
    num_mem: int

    def __init__(self, name, members):
        super(JointBase, self).__init__(name)
        self._init_check(members)
        self._beams = Beams(members)
        for m in members:
            m._ifc_elem = None

    def _init_check(self, members):
        if self.num_mem != len(members):
            raise ValueError(f"This joint only supports {self.num_mem} members")

        mem_types = [m for m in self.mem_types]

        for m in members:
            if m.member_type in mem_types:
                mem_types.pop(mem_types.index(m.member_type))

        if len(mem_types) != 0:
            raise ValueError(f"Not all Pre-requisite member types {self.mem_types} are found for JointB")

    def _cut_intersecting_member(self, mem_base, mem_incoming):
        """

        :param mem_base:
        :param mem_incoming:
        :type mem_base: ada.Beam
        :type mem_incoming: ada.Beam
        """
        h_vec = np.array(mem_incoming.up) * mem_incoming.section.h * 2
        p1, p2 = mem_base.bbox
        p1 = np.array(p1) - h_vec
        p2 = np.array(p2) + h_vec
        mem_incoming.add_penetration(PrimBox(f"{self.name}_neg", p1, p2))


class JointB(JointBase):
    mem_types = ["Column", "Girder", "Girder"]
    beamtypes = ["IG", "IG", "IG"]
    num_mem = 3

    def __init__(self, name, members):
        super(JointB, self).__init__(name, members)

        column = None
        gi1 = None
        gi2 = None
        for m in members:
            if m.member_type == "Column":
                column = m
            elif m.member_type == "Girder" and gi1 is None:
                gi1 = m
            else:
                gi2 = m

        self._cut_intersecting_member(column, gi1)
        self._cut_intersecting_member(column, gi2)

        center, s, t = beam_cross_check(column, gi1)
        self.adjust_column(column, gi1, gi2, s)

    def adjust_column(self, column, gi1, gi2, s):
        """

        :param column:
        :param gi1:
        :param gi2:
        :param s:
        :type column: ada.Beam
        :type gi1: ada.Beam
        :type gi2: ada.Beam
        :return:
        """

        dist = max(gi1.section.h, gi2.section.h) / 2
        xvec = column.xvec

        adjust_vec = 1 if s == 1.0 else -1
        adjust_col_end = adjust_vec * xvec * dist
        if adjust_vec == 1:
            if column.e2 is None:
                column.e2 = adjust_col_end
            else:
                column.e2 += adjust_col_end
        else:
            if column.e1 is None:
                column.e1 = adjust_col_end
            else:
                column.e1 += adjust_col_end



import base64
import io
import json
import operator
import urllib.request
import zipfile

import requests

urlopen = urllib.request.urlopen
deserializers = dict(ifc4="Ifc4 (Streaming)", ifc2x3="Ifc2x3tc1 (Streaming)")


class BimServerApi:
    """
    A minimal BIMserver.org API client. Interfaces are obtained from the server
    and can be retrieved as attributes from an API instance. The interfaces
    expose their methods as functions with keyword arguments.

    Example:
    import bimserver
    client = bimserver.api(server_address, username, password)
    client.Bimsie1ServiceInterface.addProject(projectName="My new project")
    """

    class interface:
        def __init__(self, api, name):
            self.api, self.name = api, name

        def make_request(self, method, **kwargs):
            request = urlopen(
                self.api.url,
                data=json.dumps(
                    dict(
                        {
                            "request": {
                                "interface": self.name,
                                "method": method,
                                "parameters": kwargs,
                            }
                        },
                        **({"token": self.api.token} if self.api.token else {}),
                    )
                ).encode("utf-8"),
            )
            response = json.loads(request.read().decode("utf-8"))
            exception = response.get("response", {}).get("exception", None)
            if exception:
                raise Exception(exception["message"])
            else:
                return response["response"]["result"]

        def __getattr__(self, method):
            return lambda **kwargs: self.make_request(method, **kwargs)

    token = None
    interfaces = None

    def __init__(self, hostname, username=None, password=None):
        self.url = "%s/json" % hostname
        if not hostname.startswith("http://") and not hostname.startswith("https://"):
            self.url = "http://%s" % self.url

        self.interfaces = set(
            map(
                operator.itemgetter("simpleName"),
                self.MetaInterface.getServiceInterfaces(),
            )
        )
        self.interfaces.add("generateRevisionDownloadUrl")

        self.version = "1.4" if "Bimsie1AuthInterface" in self.interfaces else "1.5"

        if username is not None and password is not None:
            auth_interface = getattr(self, "Bimsie1AuthInterface", getattr(self, "AuthInterface"))
            self.token = auth_interface.login(username=username, password=password)

    def __getattr__(self, interface):
        if self.interfaces is not None and interface not in self.interfaces:

            # Some form of compatibility:
            if self.version == "1.4" and not interface.startswith("Bimsie1"):
                return self.__getattr__("Bimsie1" + interface)
            elif self.version == "1.5" and interface.startswith("Bimsie1"):
                return self.__getattr__(interface[len("Bimsie1") :])

            raise AttributeError("'%s' is does not name a valid interface on this server" % interface)
        return BimServerApi.interface(self, interface)


def download_extract_zip(url):
    """
    Download a ZIP file and extract its contents in memory
    yields (filename, file-like object) pairs

    code from: https://techoverflow.net/2018/01/16/downloading-reading-a-zip-file-in-memory-using-python/

    """
    response = requests.get(url)
    with zipfile.ZipFile(io.BytesIO(response.content)) as thezip:
        for zipinfo in thezip.infolist():
            with thezip.open(zipinfo) as thefile:
                yield zipinfo.filename, thefile


class BimServerConnect:
    """
    BimServerConnect extends the bimserver json api and allows for simple pull/push operations on the server.

    ServiceInterface

    https://github.com/opensourceBIM/BIMserver/blob/8bc7413132ed934d13d4759ad5aadc1e9692f78b/PluginBase/src/org/bimserver/shared/interfaces/ServiceInterface.java

    :param bimserver_url:
    :param username: Valid username
    :param password: Password
    :param assembly:
    :param allow_create_project: Allow creation of projects
    :type assembly: ada.Assembly
    """

    def __init__(self, bimserver_url, username, password, assembly, allow_create_project=True):

        self._bimserver_url = bimserver_url
        self.client = BimServerApi(bimserver_url, username, password)
        self.assembly = assembly
        self._allow_create_project = allow_create_project

    def _get_project(self, project_name, allow_create_project):

        projects = self.client.ServiceInterface.getProjectsByName(name=project_name)

        if len(projects) > 1:
            raise ValueError("More than 1 project was found")
        elif len(projects) == 1:
            print(f'Project "{project_name}" was found!')
            project = projects[0]
        else:
            if allow_create_project is False:
                raise ValueError(f'The project "{project_name}" was not found')
            print(f'Creating new project with name "{project_name}"')
            project = self.client.ServiceInterface.addProject(projectName=project_name, schema="ifc4")
        return project

    def push(self, project_name, comment, merge=False, sync=False):
        """
        Push to BimServer

        :param project_name:
        :param comment: Commit message
        :param merge:
        :param sync:
        """

        deserializer_id = self.client.ServiceInterface.getDeserializerByName(
            deserializerName=deserializers["ifc4"]
        ).get("oid")

        # Write to IFC file and read file into memory
        ifc_f = "temp/exported.ifc"
        self.assembly.to_ifc(ifc_f)

        with open(ifc_f, "rb") as f:
            ifc_data = f.read()

        project = self._get_project(project_name, self._allow_create_project)
        project_id = project.get("oid")
        sync = "false" if sync is False else "true"
        merge = "false" if merge is False else "true"
        res = self.client.ServiceInterface.checkinSync(
            poid=project_id,
            comment=comment,
            deserializerOid=deserializer_id,
            fileSize=len(ifc_data),
            fileName=ifc_f,
            data=base64.b64encode(ifc_data).decode("utf-8"),
            sync=sync,
            merge=merge,
        )
        print(res)
        if "Error" in res["title"]:
            raise ValueError(res["title"])

    def pull(self, project_name, checkout):
        """

        :param project_name:
        :param checkout:
        :return:
        """
        import json

        project = self._get_project(project_name, self._allow_create_project)
        serializer = self.client.ServiceInterface.getSerializerByName(serializerName="Ifc4 (Streaming)")
        serializer_id = serializer.get("oid")

        if checkout is True:
            # Insert checkout code here
            pass

        roid = project.get("lastRevisionId")
        topicId = self.client.ServiceInterface.download(
            roids=[roid],
            serializerOid=serializer_id,
            query=json.dumps({}),
            sync="false",
        )

        download_url = f"{self._bimserver_url}/download?token={self.client.token}&zip=on&topicId={topicId}"
        for filename, file_obj in download_extract_zip(download_url):
            with open(filename, "wb") as d:
                d.write(file_obj.read())
                if self.assembly is not None:
                    self.assembly.read_ifc(filename)

        print("complete")


def bs_get_project(project_name, bimserver_url, username, password, allow_create_project=True, schema="ifc4"):
    """
    For manual interaction with bimserver

    :param project_name:
    :param bimserver_url:
    :param username:
    :param password:
    :param allow_create_project:
    :param schema:
    :return:
    """
    client = BimServerApi(bimserver_url, username, password)
    projects = client.ServiceInterface.getProjectsByName(name=project_name)

    if len(projects) > 1:
        raise ValueError("More than 1 project was found")
    elif len(projects) == 1:
        print(f'Project "{project_name}" was found!')
        project = projects[0]
    else:
        if allow_create_project is False:
            raise ValueError(f'The project "{project_name}" was not found')
        print(f'Creating new project with name "{project_name}"')
        project = client.ServiceInterface.addProject(projectName=project_name, schema=schema)
    return project


def bs_push(
    ifc_file,
    bimserver_url,
    username,
    password,
    project_name,
    comment,
    merge=False,
    sync=False,
    allow_create_project=True,
    schema="ifc4",
):
    """
    For manual interaction with bimserver

    :param ifc_file:
    :param bimserver_url:
    :param username:
    :param password:
    :param project_name:
    :param comment:
    :param merge:
    :param sync:
    :param allow_create_project:
    :param schema:
    :return:
    """
    client = BimServerApi(bimserver_url, username, password)

    deserializer_id = client.ServiceInterface.getDeserializerByName(deserializerName=deserializers[schema]).get("oid")

    # Write to IFC file and read file into memory

    with open(ifc_file, "rb") as f:
        ifc_data = f.read()

    project = bs_get_project(project_name, bimserver_url, username, password, allow_create_project, schema)
    project_id = project.get("oid")
    sync = "false" if sync is False else "true"
    merge = "false" if merge is False else "true"
    res = client.ServiceInterface.checkinSync(
        poid=project_id,
        comment=comment,
        deserializerOid=deserializer_id,
        fileSize=len(ifc_data),
        fileName=ifc_file,
        data=base64.b64encode(ifc_data).decode("utf-8"),
        sync=sync,
        merge=merge,
    )
    print(res)
    if "Error" in res["title"]:
        raise ValueError(res["title"])



from ada import Assembly, Node, Part
from ada.core.containers import Nodes
from ada.fem import FEM, Elem
from ada.fem.containers import FemElements


def read_assembly_from_cache(h5_filename):
    import h5py

    f = h5py.File(h5_filename, "r")
    info = f["INFO"].attrs
    a = Assembly(info["NAME"])
    parts = f["PARTS"]
    for name, p in parts.items():
        p = get_part_from_cache(name, p)
        a.add_part(p)

    return a


def get_part_from_cache(name, cache_p):
    p = Part(name)
    if "FEM" in cache_p.keys():
        p._fem = get_fem_from_cache(cache_p["FEM"])
    return p


def get_fem_from_cache(cache_fem):
    node_groups = cache_fem["NODES"]
    points = node_groups[()]
    point_ids = node_groups.attrs["IDS"][()]
    fem = FEM(cache_fem.attrs["NAME"])
    nodes = Nodes([Node(points[int(pid - 1)], pid) for pid in point_ids], parent=fem)
    fem._nodes = nodes
    elements = []
    for eltype, mesh in cache_fem["MESH"].items():
        el_ids = mesh["ELEMENTS"][()]
        elements += [Elem(el_id[0], [nodes.from_id(eli) for eli in el_id[1:]], eltype) for el_id in el_ids]
    fem._elements = FemElements(elements, fem)

    return fem



from dataclasses import dataclass

import numpy as np
from pythreejs import Group

import ada.fem

from .common import (
    get_bounding_box,
    get_edges_from_fem,
    get_faces_from_fem,
    get_vertices_from_fem,
)
from .threejs_geom import edges_to_mesh, faces_to_mesh, vertices_to_mesh


@dataclass
class ViewItem:
    fem: ada.fem.FEM
    vertices: np.array
    edges: np.array
    faces: np.array


class BBox:
    max: list
    min: list
    center: list


class FemRenderer:
    def __init__(self):
        self._view_items = []
        self._meshes = []

        # the group of 3d and 2d objects to render
        self._displayed_pickable_objects = Group()

    def add_fem(self, fem):
        vertices, faces, edges = get_vertices_from_fem(fem), get_faces_from_fem(fem), get_edges_from_fem(fem)
        self._view_items.append(ViewItem(fem, vertices, edges, faces))

    def _view_to_mesh(
        self, vt, face_colors=None, vertex_colors=(8, 8, 8), edge_color=(8, 8, 8), edge_width=1, vertex_width=1
    ):
        """

        :param vt:
        :type vt: ViewItem
        :return:
        """
        fem = vt.fem
        vertices = vt.vertices
        edges = vt.edges
        faces = vt.faces

        vertices_m = vertices_to_mesh(f"{fem.name}_vertices", vertices, vertex_colors, vertex_width)
        edges_m = edges_to_mesh(f"{fem.name}_edges", vertices, edges, edge_color=edge_color, linewidth=edge_width)
        face_geom, faces_m = faces_to_mesh(f"{fem.name}_faces", vertices, faces, colors=face_colors)

        return vertices_m, edges_m, faces_m

    def get_bounding_box(self):
        bounds = np.asarray([get_bounding_box(m) for m in self._meshes], dtype="float32")
        mi, ma = np.min(bounds, 0), np.max(bounds, 0)
        center = (mi + ma) / 2
        return mi, ma, center
